import { SocketError } from "./errors";

export type SocketHeartbeatMessage = ["heartbeat", number];
export type SocketDataMessage<T = unknown> = ["data", T];

/**
 * Creates a socket data message object with the given data.
 *
 * @param data - The data to be included in the message. Can be of any type.
 * @returns An object representing the socket data message with a type of "message" and the provided data.
 */
function wrap(data: unknown): SocketDataMessage {
  return ["data", data];
}

/**
 * Extracts the data from a SocketDataMessage.
 *
 * @param message - The socket data message to extract data from.
 * @returns The data contained in the message.
 */
function unwrap<T = unknown>(message: SocketDataMessage<T>): T {
  return message[1];
}

/**
 * Deserializes a JSON string into a `SocketMessage` object.
 *
 * @param data - The JSON string to be deserialized.
 * @returns The deserialized `SocketMessage` object.
 * @throws {ManagedSocketError} If the JSON string cannot be parsed or if the message shape is invalid.
 */
function deserialize<T>(
  data: string,
): SocketDataMessage<T> | SocketHeartbeatMessage {
  try {
    const parsedData = JSON.parse(data);

    if (typeof parsedData === "number") {
      return ["heartbeat", parsedData];
    }

    if (
      parsedData !== null &&
      Array.isArray(parsedData) &&
      parsedData.length === 2 &&
      parsedData[0] === "data"
    ) {
      const data = parsedData[1] as T;
      return ["data", data];
    }

    throw new SocketError("Failed to parse internal message shape");
  } catch (error) {
    throw new SocketError(
      "Failed to parse incoming message shape: Failed to deserialize",
    );
  }
}

export { wrap, unwrap, deserialize };
